package com.sun.jndi.ldap;

import java.io.UnsupportedEncodingException;
import java.util.Locale;
import javax.naming.directory.Attribute;
import javax.naming.directory.BasicAttributes;
import javax.naming.directory.Attributes;
import java.io.StreamCorruptedException;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.Enumeration;
import javax.naming.InvalidNameException;
import java.util.Vector;
import javax.naming.Name;

public final class LdapName implements Name
{
    private transient String unparsed;
    private transient Vector<LdapName.Rdn> rdns;
    private transient boolean valuesCaseSensitive;
    static final long serialVersionUID = -1595520034788997356L;
    
    public LdapName(String unparsed) throws InvalidNameException {
        this.valuesCaseSensitive = false;
        this.unparsed = unparsed;
        this.parse();
    }
    
    private LdapName(String unparsed, Vector<LdapName.Rdn> vector) {
        this.valuesCaseSensitive = false;
        this.unparsed = unparsed;
        this.rdns = (Vector)vector.clone();
    }
    
    private LdapName(String unparsed, Vector<LdapName.Rdn> vector, int i, int n) {
        this.valuesCaseSensitive = false;
        this.unparsed = unparsed;
        this.rdns = new Vector();
        for (; i < n; ++i) {
            this.rdns.addElement(vector.elementAt(i));
        }
    }
    
    @Override
    public Object clone() {
        return new LdapName(this.unparsed, this.rdns);
    }
    
    @Override
    public String toString() {
        if (this.unparsed != null) {
            return this.unparsed;
        }
        StringBuffer buffer = new StringBuffer();
        for (int i = this.rdns.size() - 1; i >= 0; --i) {
            if (i < this.rdns.size() - 1) {
                buffer.append(',');
            }
            buffer.append((LdapName.Rdn)this.rdns.elementAt(i));
        }
        return this.unparsed = new String(buffer);
    }
    
    @Override
    public boolean equals(Object o) {
        return o instanceof LdapName && this.compareTo(o) == 0;
    }
    
    @Override
    public int compareTo(Object o) {
        LdapName ldapName = (LdapName)o;
        if (o == this || (this.unparsed != null && this.unparsed.equals(ldapName.unparsed))) {
            return 0;
        }
        int min = Math.min(this.rdns.size(), ldapName.rdns.size());
        int i = 0;
        while (i < min) {
            int compareTo = ((LdapName.Rdn)this.rdns.elementAt(i)).compareTo((LdapName.Rdn)ldapName.rdns.elementAt(i));
            if (compareTo != 0) {
                return compareTo;
            }
            ++i;
        }
        return this.rdns.size() - ldapName.rdns.size();
    }
    
    @Override
    public int hashCode() {
        int n = 0;
        for (int i = 0; i < this.rdns.size(); ++i) {
            n += ((LdapName.Rdn)this.rdns.elementAt(i)).hashCode();
        }
        return n;
    }
    
    @Override
    public int size() {
        return this.rdns.size();
    }
    
    @Override
    public boolean isEmpty() {
        return this.rdns.isEmpty();
    }
    
    @Override
    public Enumeration<String> getAll() {
        return new Enumeration<String>() {
            
            @Override
            public boolean hasMoreElements() {
                return enum_.hasMoreElements();
            }
            
            @Override
            public String nextElement() {
                return ((LdapName.Rdn)enum_.nextElement()).toString();
            }
        };
    }
    
    @Override
    public String get(int index) {
        return ((LdapName.Rdn)this.rdns.elementAt(index)).toString();
    }
    
    @Override
    public Name getPrefix(int n) {
        return new LdapName(null, this.rdns, 0, n);
    }
    
    @Override
    public Name getSuffix(int n) {
        return new LdapName(null, this.rdns, n, this.rdns.size());
    }
    
    @Override
    public boolean startsWith(Name name) {
        int size = this.rdns.size();
        int size2 = name.size();
        return size >= size2 && this.matches(0, size2, name);
    }
    
    @Override
    public boolean endsWith(Name name) {
        int size = this.rdns.size();
        int size2 = name.size();
        return size >= size2 && this.matches(size - size2, size, name);
    }
    
    public void setValuesCaseSensitive(boolean valuesCaseSensitive) {
        this.toString();
        this.rdns = null;
        try {
            this.parse();
        }
        catch (InvalidNameException ex) {
            throw new IllegalStateException("Cannot parse name: " + this.unparsed);
        }
        this.valuesCaseSensitive = valuesCaseSensitive;
    }
    
    private boolean matches(int i, int n, Name name) {
        while (i < n) {
            LdapName.Rdn rdn;
            if ((name instanceof LdapName)) {
                rdn = (LdapName.Rdn)((LdapName)name).rdns.elementAt(i - i);
            }
            else {
                String value = name.get(i - i);
                try {
                    rdn = new LdapName.DnParser(value, this.valuesCaseSensitive).getRdn();
                }
                catch (InvalidNameException ex) {
                    return false;
                }
            }
            if (!rdn.equals(this.rdns.elementAt(i))) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    @Override
    public Name addAll(Name name) throws InvalidNameException {
        return this.addAll(this.size(), name);
    }
    
    @Override
    public Name addAll(int n, Name name) throws InvalidNameException {
        if ((name instanceof LdapName)) {
            LdapName ldapName = (LdapName)name;
            for (int i = 0; i < ldapName.rdns.size(); ++i) {
                this.rdns.insertElementAt(ldapName.rdns.elementAt(i), n++);
            }
        }
        else {
            Enumeration<String> all = name.getAll();
            while (all.hasMoreElements()) {
                this.rdns.insertElementAt(new LdapName.DnParser((String)all.nextElement(), this.valuesCaseSensitive).getRdn(), n++);
            }
        }
        this.unparsed = null;
        return this;
    }
    
    @Override
    public Name add(String s) throws InvalidNameException {
        return this.add(this.size(), s);
    }
    
    @Override
    public Name add(int index, String s) throws InvalidNameException {
        this.rdns.insertElementAt(new LdapName.DnParser(s, this.valuesCaseSensitive).getRdn(), index);
        this.unparsed = null;
        return this;
    }
    
    @Override
    public Object remove(int index) throws InvalidNameException {
        String value = this.get(index);
        this.rdns.removeElementAt(index);
        this.unparsed = null;
        return value;
    }
    
    private void parse() throws InvalidNameException {
        this.rdns = new LdapName.DnParser(this.unparsed, this.valuesCaseSensitive).getDn();
    }
    
    private static boolean isWhitespace(char c) {
        return c == ' ' || c == '\r';
    }
    
    public static String escapeAttributeValue(Object o) {
        return LdapName.TypeAndValue.escapeValue(o);
    }
    
    public static Object unescapeAttributeValue(String s) {
        return LdapName.TypeAndValue.unescapeValue(s);
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        objectOutputStream.writeObject(this.toString());
        objectOutputStream.writeBoolean(this.valuesCaseSensitive);
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        this.unparsed = (String)objectInputStream.readObject();
        this.valuesCaseSensitive = objectInputStream.readBoolean();
        try {
            this.parse();
        }
        catch (InvalidNameException ex) {
            throw new StreamCorruptedException("Invalid name: " + this.unparsed);
        }
    }
    
    static boolean access$000(char c) {
        return isWhitespace(c);
    }
    
    static class DnParser
    {
        private final String name;
        private final char[] chars;
        private final int len;
        private int cur;
        private boolean valuesCaseSensitive;
        
        DnParser(String name, boolean valuesCaseSensitive) throws InvalidNameException {
            this.cur = 0;
            this.name = name;
            this.len = name.length();
            this.chars = name.toCharArray();
            this.valuesCaseSensitive = valuesCaseSensitive;
        }
        
        Vector<LdapName.Rdn> getDn() throws InvalidNameException {
            this.cur = 0;
            Vector<LdapName.Rdn> vector = new Vector(this.len / 3 + 10);
            if (this.len == 0) {
                return vector;
            }
            vector.addElement(this.parseRdn());
            while (this.cur < this.len) {
                if (this.chars[this.cur] == ',' || this.chars[this.cur] == ';') {
                    ++this.cur;
                    vector.insertElementAt(this.parseRdn(), 0);
                }
                else {
                    throw new InvalidNameException("Invalid name: " + this.name);
                }
            }
            return vector;
        }
        
        LdapName.Rdn getRdn() throws InvalidNameException {
            LdapName.Rdn rdn = this.parseRdn();
            if (this.cur < this.len) {
                throw new InvalidNameException("Invalid RDN: " + this.name);
            }
            return rdn;
        }
        
        private LdapName.Rdn parseRdn() throws InvalidNameException {
            LdapName.Rdn rdn = new LdapName.Rdn();
            while (this.cur < this.len) {
                this.consumeWhitespace();
                String attrType = this.parseAttrType();
                this.consumeWhitespace();
                if (this.cur >= this.len || this.chars[this.cur] != '=') {
                    throw new InvalidNameException("Invalid name: " + this.name);
                }
                ++this.cur;
                this.consumeWhitespace();
                String attrValue = this.parseAttrValue();
                this.consumeWhitespace();
                rdn.add(new LdapName.TypeAndValue(attrType, attrValue, this.valuesCaseSensitive));
                if (this.cur < this.len) {
                    if (this.chars[this.cur] != '+') {
                        break;
                    }
                    ++this.cur;
                }
                else {
                    break;
                }
            }
            return rdn;
        }
        
        private String parseAttrType() throws InvalidNameException {
            int cur = this.cur;
            while (this.cur < this.len) {
                int ch = this.chars[this.cur];
                if (Character.isLetterOrDigit((char)ch) || ch == '.' || ch == '-' || ch == ' ') {
                    ++this.cur;
                }
                else {
                    break;
                }
            }
            while (this.cur > cur && this.chars[this.cur - 1] == ' ') {
                --this.cur;
            }
            if (cur == this.cur) {
                throw new InvalidNameException("Invalid name: " + this.name);
            }
            return new String(this.chars, cur, this.cur - cur);
        }
        
        private String parseAttrValue() throws InvalidNameException {
            if (this.cur < this.len && this.chars[this.cur] == '#') {
                return this.parseBinaryAttrValue();
            }
            if (this.cur < this.len && this.chars[this.cur] == '\"') {
                return this.parseQuotedAttrValue();
            }
            return this.parseStringAttrValue();
        }
        
        private String parseBinaryAttrValue() throws InvalidNameException {
            int cur = this.cur;
            ++this.cur;
            while (this.cur < this.len && Character.isLetterOrDigit(this.chars[this.cur])) {
                ++this.cur;
            }
            return new String(this.chars, cur, this.cur - cur);
        }
        
        private String parseQuotedAttrValue() throws InvalidNameException {
            int cur = this.cur;
            ++this.cur;
            while (this.cur < this.len && this.chars[this.cur] != '\"') {
                if (this.chars[this.cur] == '\\') {
                    ++this.cur;
                }
                ++this.cur;
            }
            if (this.cur >= this.len) {
                throw new InvalidNameException("Invalid name: " + this.name);
            }
            ++this.cur;
            return new String(this.chars, cur, this.cur - cur);
        }
        
        private String parseStringAttrValue() throws InvalidNameException {
            int cur = this.cur;
            int cur2 = -1;
            while (this.cur < this.len && !this.atTerminator()) {
                if (this.chars[this.cur] == '\\') {
                    ++this.cur;
                    cur2 = this.cur;
                }
                ++this.cur;
            }
            if (this.cur > this.len) {
                throw new InvalidNameException("Invalid name: " + this.name);
            }
            int cur3;
            for (cur3 = this.cur; cur3 > cur && LdapName.access$000(this.chars[cur3 - 1]) && cur2 != cur3 - 1; --cur3) {}
            return new String(this.chars, cur, cur3 - cur);
        }
        
        private void consumeWhitespace() {
            while (this.cur < this.len && LdapName.access$000(this.chars[this.cur])) {
                ++this.cur;
            }
        }
        
        private boolean atTerminator() {
            return this.cur < this.len && (this.chars[this.cur] == ',' || this.chars[this.cur] == ';' || this.chars[this.cur] == '+');
        }
    }
    
    static class Rdn
    {
        private final Vector<LdapName.TypeAndValue> tvs;
        
        Rdn() {
            this.tvs = new Vector();
        }
        
        void add(LdapName.TypeAndValue obj) {
            int i = 0;
            while (i < this.tvs.size()) {
                int compareTo = obj.compareTo(this.tvs.elementAt(i));
                if (compareTo == 0) {
                    return;
                }
                if (compareTo < 0) {
                    break;
                }
                ++i;
            }
            this.tvs.insertElementAt(obj, i);
        }
        
        @Override
        public String toString() {
            StringBuffer buffer = new StringBuffer();
            for (int i = 0; i < this.tvs.size(); ++i) {
                if (i > 0) {
                    buffer.append('+');
                }
                buffer.append(this.tvs.elementAt(i));
            }
            return new String(buffer);
        }
        
        @Override
        public boolean equals(Object o) {
            return o instanceof LdapName.Rdn && this.compareTo(o) == 0;
        }
        
        public int compareTo(Object o) {
            LdapName.Rdn rdn = (LdapName.Rdn)o;
            int min = Math.min(this.tvs.size(), rdn.tvs.size());
            int i = 0;
            while (i < min) {
                int compareTo = ((LdapName.TypeAndValue)this.tvs.elementAt(i)).compareTo(rdn.tvs.elementAt(i));
                if (compareTo != 0) {
                    return compareTo;
                }
                ++i;
            }
            return this.tvs.size() - rdn.tvs.size();
        }
        
        @Override
        public int hashCode() {
            int n = 0;
            for (int i = 0; i < this.tvs.size(); ++i) {
                n += ((LdapName.TypeAndValue)this.tvs.elementAt(i)).hashCode();
            }
            return n;
        }
        
        Attributes toAttributes() {
            BasicAttributes basicAttributes = new BasicAttributes(true);
            for (int i = 0; i < this.tvs.size(); ++i) {
                LdapName.TypeAndValue typeAndValue = (LdapName.TypeAndValue)this.tvs.elementAt(i);
                Attribute value;
                if ((value = basicAttributes.get(typeAndValue.getType())) == null) {
                    basicAttributes.put(typeAndValue.getType(), typeAndValue.getUnescapedValue());
                }
                else {
                    value.add(typeAndValue.getUnescapedValue());
                }
            }
            return basicAttributes;
        }
    }
    
    static class TypeAndValue
    {
        private final String type;
        private final String value;
        private final boolean binary;
        private final boolean valueCaseSensitive;
        private String comparable;
        
        TypeAndValue(String type, String value, boolean valueCaseSensitive) {
            this.comparable = null;
            this.type = type;
            this.value = value;
            this.binary = value.startsWith("#");
            this.valueCaseSensitive = valueCaseSensitive;
        }
        
        @Override
        public String toString() {
            return this.type + "=" + this.value;
        }
        
        public int compareTo(Object o) {
            LdapName.TypeAndValue typeAndValue = (LdapName.TypeAndValue)o;
            int compareToIgnoreCase = this.type.compareToIgnoreCase(typeAndValue.type);
            if (compareToIgnoreCase != 0) {
                return compareToIgnoreCase;
            }
            if (this.value.equals(typeAndValue.value)) {
                return 0;
            }
            return this.getValueComparable().compareTo(typeAndValue.getValueComparable());
        }
        
        @Override
        public boolean equals(Object o) {
            if (!(o instanceof LdapName.TypeAndValue)) {
                return false;
            }
            LdapName.TypeAndValue typeAndValue = (LdapName.TypeAndValue)o;
            return this.type.equalsIgnoreCase(typeAndValue.type) && (this.value.equals(typeAndValue.value) || this.getValueComparable().equals(typeAndValue.getValueComparable()));
        }
        
        @Override
        public int hashCode() {
            return this.type.toUpperCase(Locale.ENGLISH).hashCode() + this.getValueComparable().hashCode();
        }
        
        String getType() {
            return this.type;
        }
        
        Object getUnescapedValue() {
            return unescapeValue(this.value);
        }
        
        private String getValueComparable() {
            if (this.comparable != null) {
                return this.comparable;
            }
            if (this.binary) {
                this.comparable = this.value.toUpperCase(Locale.ENGLISH);
            }
            else {
                this.comparable = (String)unescapeValue(this.value);
                if (!this.valueCaseSensitive) {
                    this.comparable = this.comparable.toUpperCase(Locale.ENGLISH);
                }
            }
            return this.comparable;
        }
        
        static String escapeValue(Object o) {
            return (o instanceof byte[]) ? escapeBinaryValue((byte[])(byte[])o) : escapeStringValue((String)o);
        }
        
        private static String escapeStringValue(String s) {
            char[] charArray = s.toCharArray();
            StringBuffer buffer = new StringBuffer(2 * s.length());
            int n;
            for (n = 0; n < charArray.length && LdapName.access$000(charArray[n]); ++n) {}
            int n2;
            for (n2 = charArray.length - 1; n2 >= 0 && LdapName.access$000(charArray[n2]); --n2) {}
            for (int i = 0; i < charArray.length; ++i) {
                int n3 = charArray[i];
                if (i < n || i > n2 || ",=+<>#;\"\\".indexOf(n3) >= 0) {
                    buffer.append('\\');
                }
                buffer.append((char)n3);
            }
            return new String(buffer);
        }
        
        private static String escapeBinaryValue(byte[] array) {
            StringBuffer buffer = new StringBuffer(1 + 2 * array.length);
            buffer.append("#");
            for (int i = 0; i < array.length; ++i) {
                int n = array[i];
                buffer.append(Character.forDigit(0xF & n >>> 4, 16));
                buffer.append(Character.forDigit(0xF & n, 16));
            }
            return new String(buffer).toUpperCase(Locale.ENGLISH);
        }
        
        static Object unescapeValue(String str) {
            char[] charArray;
            int i;
            int length;
            for (charArray = str.toCharArray(), i = 0, length = charArray.length; i < length && LdapName.access$000(charArray[i]); ++i) {}
            while (i < length && LdapName.access$000(charArray[length - 1])) {
                --length;
            }
            if (length != charArray.length && i < length && charArray[length - 1] == '\\') {
                ++length;
            }
            if (i >= length) {
                return "";
            }
            if (charArray[i] == '#') {
                return decodeHexPairs(charArray, ++i, length);
            }
            if (charArray[i] == '\"' && charArray[length - 1] == '\"') {
                ++i;
                --length;
            }
            StringBuffer buffer = new StringBuffer(length - i);
            int n = -1;
            for (; i < length; ++i) {
                if (charArray[i] == '\\' && i + 1 < length) {
                    if (!Character.isLetterOrDigit(charArray[i + 1])) {
                        ++i;
                        buffer.append(charArray[i]);
                        n = i;
                    }
                    else {
                        byte[] utf8Octets = getUtf8Octets(charArray, i, length);
                        if (utf8Octets.length > 0) {
                            try {
                                buffer.append(new String(utf8Octets, "UTF8"));
                            }
                            catch (UnsupportedEncodingException ex) {}
                            i += utf8Octets.length * 3 - 1;
                        }
                        else {
                            throw new IllegalArgumentException("Not a valid attribute string value:" + str + ", improper usage of backslash");
                        }
                    }
                }
                else {
                    buffer.append(charArray[i]);
                }
            }
            int length2 = buffer.length();
            if (LdapName.access$000(buffer.charAt(length2 - 1)) && n != length - 1) {
                buffer.setLength(length2 - 1);
            }
            return new String(buffer);
        }
        
        private static byte[] decodeHexPairs(char[] value, int n, int n2) {
            byte[] array = new byte[(n2 - n) / 2];
            int n3 = 0;
            while (n + 1 < n2) {
                int digit = Character.digit(value[n], 16);
                int digit2 = Character.digit(value[n + 1], 16);
                if (digit >= 0) {
                    if (digit2 < 0) {
                        break;
                    }
                    array[n3] = (byte)((digit << 4) + digit2);
                    n += 2;
                    ++n3;
                }
                else {
                    break;
                }
            }
            if (n != n2) {
                throw new IllegalArgumentException("Illegal attribute value: #" + new String(value));
            }
            return array;
        }
        
        private static byte[] getUtf8Octets(char[] array, int n, int n2) {
            byte[] array2 = new byte[(n2 - n) / 3];
            int n3 = 0;
            while (n + 2 < n2 && array[n++] == '\\') {
                int digit = Character.digit(array[n++], 16);
                int digit2 = Character.digit(array[n++], 16);
                if (digit >= 0) {
                    if (digit2 < 0) {
                        break;
                    }
                    array2[n3++] = (byte)((digit << 4) + digit2);
                }
                else {
                    break;
                }
            }
            if (n3 == array2.length) {
                return array2;
            }
            byte[] array3 = new byte[n3];
            System.arraycopy(array2, 0, array3, 0, n3);
            return array3;
        }
    }
}
